# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import click

from threading import Lock
from typing import cast, Iterable, Type
from uuid import UUID

from odc.geo import Geometry, CRS
from datacube import Datacube
from datacube.model import Product, Dataset

from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.index.api import OWSAbstractIndex, OWSAbstractIndexDriver, LayerSignature, LayerExtent, TimeSearchTerm
from .product_ranges import create_range_entry as create_range_entry_impl, get_ranges as get_ranges_impl
from .mv_index import MVSelectOpts, mv_search
from datacube_ows.index.sql import run_sql


class OWSPostgresIndex(OWSAbstractIndex):
    name: str = "postgres"

    # method to delete obsolete schemas etc.
    def cleanup_schema(self, dc: Datacube):
        self._run_sql(dc, "ows_schema/cleanup")

    # Schema creation method
    def create_schema(self, dc: Datacube):
        click.echo("Creating/updating schema and tables...")
        self._run_sql(dc, "ows_schema/create")
        click.echo("Creating/updating materialised views...")
        self._run_sql(dc, "extent_views/create")
        click.echo("Setting ownership of materialised views...")
        self._run_sql(dc, "extent_views/grants/refresh_owner")

    # Permission management method
    def grant_perms(self, dc: Datacube, role: str, read_only: bool = False):
        if read_only:
            self._run_sql(dc, "ows_schema/grants/read_only", role=role)
            self._run_sql(dc, "extent_views/grants/read_only", role=role)
        else:
            self._run_sql(dc, "ows_schema/grants/read_write", role=role)
            self._run_sql(dc, "extent_views/grants/write_refresh", role=role)

    # Spatiotemporal index update method (e.g. refresh materialised views)
    def update_geotemporal_index(self, dc: Datacube):
        self._run_sql(dc, "extent_views/refresh")

    def create_range_entry(self, layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]) -> None:
        create_range_entry_impl(layer, cache)

    def get_ranges(self, layer: OWSNamedLayer) -> LayerExtent | None:
        return get_ranges_impl(layer)

    def ds_search(self,
                  layer: OWSNamedLayer,
                  times: Iterable[TimeSearchTerm] | None = None,
                  geom: Geometry | None = None,
                  products: Iterable[Product] | None = None
                  ) -> Iterable[Dataset]:
        return cast(Iterable[Dataset], mv_search(layer.dc.index, MVSelectOpts.DATASETS,
                                                 times=times, geom=geom, products=products))

    def dsid_search(self,
                    layer: OWSNamedLayer,
                    times: Iterable[TimeSearchTerm] | None = None,
                    geom: Geometry | None = None,
                    products: Iterable[Product] | None = None
                    ) -> Iterable[UUID]:
        return cast(Iterable[UUID], mv_search(layer.dc.index, MVSelectOpts.IDS,
                                              times=times, geom=geom, products=products))

    def count(self,
              layer: OWSNamedLayer,
              times: Iterable[TimeSearchTerm] | None = None,
              geom: Geometry | None = None,
              products: Iterable[Product] | None = None
              ) -> int:
        return cast(int, mv_search(layer.dc.index, MVSelectOpts.COUNT,
                                   times=times, geom=geom, products=products))

    def extent(self,
               layer: OWSNamedLayer,
               times: Iterable[TimeSearchTerm] | None = None,
               geom: Geometry | None = None,
               products: Iterable[Product] | None = None,
               crs: CRS | None = None
               ) -> Geometry | None:
        extent = cast(Geometry | None, mv_search(layer.dc.index, MVSelectOpts.EXTENT,
                                                 times=times, geom=geom, products=products))
        if extent is None or crs is None or crs == extent.crs:
            return extent
        else:
            return extent.to_crs(crs)

    def _run_sql(self, dc: Datacube, path: str, **params: str) -> bool:
        return run_sql(dc, self.name, path, **params)


pgdriverlock = Lock()


class OWSPostgresIndexDriver(OWSAbstractIndexDriver):
    _driver = None
    @classmethod
    def ows_index_class(cls) -> Type[OWSAbstractIndex]:
        return OWSPostgresIndex

    @classmethod
    def ows_index(cls) -> OWSAbstractIndex:
        with pgdriverlock:
            if cls._driver is None:
                cls._driver = OWSPostgresIndex()
        return cls._driver


def ows_index_driver_init():
    return OWSPostgresIndexDriver()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import math
from datetime import date, datetime, timezone
from typing import Callable
import click

import datacube
import odc.geo
import sqlalchemy.exc
from psycopg2.extras import Json
from sqlalchemy import text

from odc.geo.geom import CRS

from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.utils import get_sqlconn
from datacube_ows.index.api import CoordRange, LayerSignature, LayerExtent

_LOG = logging.getLogger(__name__)


def jsonise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    return {
        "top": bbox.top,
        "bottom": bbox.bottom,
        "left": bbox.left,
        "right": bbox.right,
    }


def create_range_entry(layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]) -> None:
    meta = LayerSignature(time_res=layer.time_resolution.value,
                          products=tuple(layer.product_names),
                          env=layer.local_env._name,
                          datasets=layer.dc.index.datasets.count(product=layer.product_names))

    click.echo(f"Postgis Updating range for layer {layer.name}")
    click.echo(f"(signature: {meta.as_json()!r})")
    conn = get_sqlconn(layer.dc)
    txn = conn.begin()
    if meta in cache:
        template = cache[meta][0]
        click.echo(f"Layer {template} has same signature - reusing")
        cache[meta].append(layer.name)
        try:
            conn.execute(text("""
            INSERT INTO ows.layer_ranges
                (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
            SELECT :layer_id, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated
            FROM ows.layer_ranges lr2
            WHERE lr2.layer = :template_id"""),
                         {
                                      "layer_id": layer.name,
                                      "template_id": template
                                   })
        except sqlalchemy.exc.IntegrityError:
            conn.execute(text("""
            UPDATE ows.layer_ranges lr1
            SET lat_min = lr2.lat_min,
                lat_max = lr2.lat_max,
                lon_min = lr2.lon_min,
                lon_max = lr2.lon_max,
                dates = lr2.dates,
                bboxes = lr2.bboxes,
                meta = lr2.meta,
                last_updated = lr2.last_updated
            FROM ows.layer_ranges lr2
            WHERE lr1.layer = :layer_id
            AND   lr2.layer = :template_id"""),
                         {
                             "layer_id": layer.name,
                             "template_id": template
                         })
    else:
        # insert empty row if one does not already exist
        conn.execute(text("""
        INSERT INTO ows.layer_ranges
        (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
        VALUES
        (:p_layer, 0, 0, 0, 0, :empty, :empty, :meta, :now)
        ON CONFLICT (layer) DO NOTHING
        """),
        {
            "p_layer": layer.name, "empty": Json(""),
            "meta": Json(meta.as_json()), "now": datetime.now(tz=timezone.utc)
        })

        prodids = [p.id for p in layer.products]

        # Set default timezone
        conn.execute(text("""set timezone to 'Etc/UTC'"""))

        # Loop over dates
        dates = set()  # Should get to here!
        if layer.time_resolution.is_solar():
          results = conn.execute(text(
              """
              select lower(dt.search_val), upper(dt.search_val),
                     lower(lon.search_val), upper(lon.search_val)
              from odc.dataset ds, odc.dataset_search_datetime dt, odc.dataset_search_num lon
              where ds.product_ref = ANY(:prodids)
              AND ds.id = dt.dataset_ref
              AND ds.id = lon.dataset_ref
              AND dt.search_key = :time
              AND lon.search_key = :lon
              """),
              {"prodids": prodids, "time": "time", "lon": "lon"})
          for result in results:
              dt1, dt2, ll, lu = result
              lon = (ll + lu) / 2
              dt = dt1 + (dt2 - dt1) / 2
              dt = dt.astimezone(timezone.utc)

              solar_day = datacube.api.query._convert_to_solar_time(dt, lon).date()
              dates.add(solar_day)
        else:
          results = conn.execute(text(
              """
              select
                    array_agg(dt.search_val)
              from odc.dataset_search_datetime dt,
                   odc.dataset ds
              WHERE ds.product_ref = ANY(:prodids)
              AND   ds.id = dt.dataset_ref
              AND   dt.search_key = 'time'
              """),
              {"prodids": prodids}
          )
          for result in results:
              for dat_ran in result[0]:
                  dates.add(dat_ran.lower)

        if layer.time_resolution.is_subday():
          date_formatter = lambda d: d.isoformat()
        else:
          date_formatter = lambda d: d.strftime("%Y-%m-%d")

        dates = sorted(dates)
        conn.execute(text("""
           UPDATE ows.layer_ranges
           SET dates = :dates
           WHERE layer= :layer_id
        """),
                   {
                       "dates": Json(list(map(date_formatter, dates))),
                       "layer_id": layer.name
                   }
        )
        # calculate bounding boxes
        # Get extent polygon from materialised views

        base_crs = CRS(layer.native_CRS)
        if base_crs not in layer.dc.index.spatial_indexes():
            click.echo(f"Native CRS for layer {layer.name} ({layer.native_CRS}) does not have a spatial index. "
                  "Using epsg:4326 for extent calculations.")
            base_crs = CRS("EPSG:4326")

        base_extent = None
        for product in layer.products:
            prod_extent = layer.dc.index.products.spatial_extent(product, base_crs)
            if base_extent is None:
                base_extent = prod_extent
            else:
                base_extent = base_extent | prod_extent
        assert base_extent is not None
        all_bboxes = bbox_projections(base_extent, layer.global_cfg.crses)

        conn.execute(text("""
        UPDATE ows.layer_ranges
        SET bboxes = :bbox,
            lat_min = :lat_min,
            lat_max = :lat_max,
            lon_min = :lon_min,
            lon_max = :lon_max
        WHERE layer = :layer_id
        """), {
            "bbox": Json(all_bboxes),
            "layer_id": layer.name,
            "lat_min": all_bboxes['EPSG:4326']['bottom'],
            "lat_max": all_bboxes['EPSG:4326']['top'],
            "lon_min": all_bboxes['EPSG:4326']['left'],
            "lon_max": all_bboxes['EPSG:4326']['right']
        })

        cache[meta] = [layer.name]

    txn.commit()
    conn.close()


def bbox_projections(starting_box: odc.geo.Geometry, crses: dict[str, odc.geo.CRS]) -> dict[str, dict[str, float]]:
   result = {}
   for crsid, crs in crses.items():
       if crs.valid_region is not None:
           test_box = starting_box.to_crs("epsg:4326")
           clipped_crs_region = (test_box & crs.valid_region)
           if clipped_crs_region.wkt == 'POLYGON EMPTY':
               continue
           clipped_crs_bbox = clipped_crs_region.to_crs(crs).boundingbox
       else:
           clipped_crs_bbox = None
       if clipped_crs_bbox is not None:
           result[crsid] = jsonise_bbox(clipped_crs_bbox)
       else:
           projbbox = starting_box.to_crs(crs).boundingbox
           result[crsid] = sanitise_bbox(projbbox)
   return result


def sanitise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    def sanitise_coordinate(coord: float, fallback: float) -> float:
        return coord if math.isfinite(coord) else fallback
    return {
        "top": sanitise_coordinate(bbox.top, float("9.999999999e99")),
        "bottom": sanitise_coordinate(bbox.bottom, float("-9.999999999e99")),
        "left": sanitise_coordinate(bbox.left, float("-9.999999999e99")),
        "right": sanitise_coordinate(bbox.right, float("9.999999999e99")),
    }


def get_ranges(layer: OWSNamedLayer) -> LayerExtent | None:
    cfg = layer.global_cfg
    conn = get_sqlconn(layer.dc)
    results = conn.execute(text("""
        SELECT *
        FROM ows.layer_ranges
        WHERE layer=:pname"""),
                           {"pname": layer.name}
                          )

    for result in results:
        conn.close()
        if layer.time_resolution.is_subday():
            dt_parser: Callable[[str], datetime | date] = lambda dts: datetime.fromisoformat(dts)
        else:
            dt_parser = lambda dts: datetime.strptime(dts, "%Y-%m-%d").date()
        times = [dt_parser(d) for d in result.dates if d is not None]
        if not times:
            return None
        return LayerExtent(
            lat=CoordRange(min=float(result.lat_min), max=float(result.lat_max)),
            lon=CoordRange(min=float(result.lon_min), max=float(result.lon_max)),
            times=times,
            bboxes=result.bboxes
        )
    return None



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os
from importlib import import_module
from itertools import chain
from typing import Any, Callable, Iterable, Optional, Sequence, cast
from urllib.parse import urlparse

import fsspec
from babel.messages import Catalog, Message
from datacube.model import Product
from datacube.utils.masking import make_mask
from flask_babel import gettext as _
from xarray import DataArray

from datacube_ows.config_toolkit import deepinherit

TYPE_CHECKING = False
if TYPE_CHECKING:
    import datacube_ows.ows_configuration.AttributionCfg
    import datacube_ows.ows_configuration.OWSConfig
    import datacube_ows.ows_configuration.OWSNamedLayer
    import datacube_ows.styles.base.StyleMask

_LOG = logging.getLogger(__name__)

RAW_CFG = None | str | int | float | bool | list["RAW_CFG"] | dict[str, "RAW_CFG"]

CFG_DICT = dict[str, RAW_CFG]

F = Callable[..., Any]


# inclusions defaulting to an empty list is dangerous, but note that it is never modified.
# If modification of inclusions is a required, a copy (ninclusions) is made and modified instead.
# pylint: disable=dangerous-default-value
def cfg_expand(cfg_unexpanded: CFG_DICT,
               cwd: str | None = None, inclusions: list[str] = []) -> CFG_DICT:
    """
    Recursively expand config inclusions.

    :param cfg_unexpanded: The unexpanded configuration object
    :param cwd: (optional) the directory relative to which json expansions will be resolved
    :param inclusions: (optional) inclusions already included (prevents infinite recursion)
    :return: The fully expanded configuration object
    """
    if cwd is None:
        cwd = os.getcwd()

    if isinstance(cfg_unexpanded, dict):
        if "include" in cfg_unexpanded:
            if cfg_unexpanded["include"] in inclusions:
                raise ConfigException("Cyclic inclusion: %s" % cfg_unexpanded["include"])
            raw_path = cast(str, cfg_unexpanded["include"])
            ninclusions: list[str] = inclusions.copy()
            ninclusions.append(cast(str, raw_path))
            # Perform expansion
            if "type" not in cfg_unexpanded or cfg_unexpanded["type"] == "json":
                # JSON Expansion
                try:
                    # Try in actual working directory
                    json_obj: Any = load_json_obj(raw_path)
                    abs_path: str = os.path.abspath(raw_path)
                    cwd = os.path.dirname(abs_path)
                # pylint: disable=broad-except
                except Exception:
                    json_obj = None
                if json_obj is None:
                    path: str = os.path.join(cwd, raw_path)
                    try:
                        # Try in inherited working directory
                        json_obj = load_json_obj(path)
                    # pylint: disable=broad-except
                    except Exception:
                        json_obj = None
                if json_obj is None:
                    raise ConfigException(f"Could not find json file {raw_path}")
                return cfg_expand(json_obj, cwd=cwd, inclusions=ninclusions)
            elif cfg_unexpanded["type"] == "python":
                # Python Expansion
                return cfg_expand(import_python_obj(raw_path), cwd=cwd, inclusions=ninclusions)
            else:
                raise ConfigException("Unsupported inclusion type: %s" % str(cfg_unexpanded["type"]))
        else:
            return {
                k: cfg_expand(cast(CFG_DICT, v), cwd=cwd, inclusions=inclusions)
                for k, v in cfg_unexpanded.items()
            }
    elif isinstance(cfg_unexpanded, Sequence) and not isinstance(cfg_unexpanded, str):
        return [cfg_expand(elem, cwd=cwd, inclusions=inclusions) for elem in cfg_unexpanded]
    else:
        return cfg_unexpanded


def get_file_loc(x: str) -> str:
    """Helper function to deal with local / remote "working directory"

    Returns the absolute pathname for a local file
    and the URL location of a remote file.
    """
    xp = urlparse(x)
    if xp.scheme in ("s3",): # NOTE: could add http/s, ...
        enable_s3 = os.environ.get("DATACUBE_OWS_CFG_ALLOW_S3", "no")
        if not enable_s3.lower() in ("yes", "true", "1", "y"):
            raise ConfigException("Please set environment variable 'DATACUBE_OWS_CFG_ALLOW_S3=YES' "
                              + "to enable OWS config from AWS S3")
        cwd = xp.scheme + "://" + xp.netloc +  xp.path.rsplit("/", 1)[0]
    elif xp.scheme:
        raise ConfigException(f"Unsupported URL scheme in config inheritance: {xp.scheme}")
    else:
        abs_path = os.path.abspath(x)
        cwd = os.path.dirname(abs_path)
    return cwd


def load_json_obj(path: str) -> RAW_CFG:
    """
    Load a json object from a file path
    :param path: The file path
    :return: The deserialised json object from the file.
    """
    with fsspec.open(path) as json_file:
        return json.load(json_file)


def import_python_obj(path: str) -> CFG_DICT:
    """Imports a python dictionary by fully-qualified path

    :param: A fully qualified python path.
    :return: a Python object, or None
    """
    mod_name, obj_name = path.rsplit('.', 1)
    try:
        mod = import_module(mod_name)
        obj = getattr(mod, obj_name)
    except (ImportError, ValueError, ModuleNotFoundError, AttributeError):
        raise ConfigException(f"Could not import python object: {path}")
    return cast(CFG_DICT, obj)


class ConfigException(Exception):
    """
    General exception for OWS Configuration issues.
    """


class OWSConfigNotReady(ConfigException):
    """
    Exception raised when someone tries to use an OWSConfigEntry that isn't fully initialised yet.
    """


class ODCInitException(ConfigException):
    """
    Exception raised when a Datacube index could not be created
    """



class OWSConfigEntry:
    """
    Base class for all configuration objects
    """
    # Parse and validate the json but don't access the database.
    def __init__(self, cfg: RAW_CFG, *args, **kwargs) -> None:
        """
        Base Class Constructor.

        Handles unready attributes for two-phase intitialisation and stows the raw configuration away.

        :param cfg: The congfiguration being parsed
        :param args:
        :param kwargs:
        """
        self._unready_attributes: set[str] = set()
        self._raw_cfg: RAW_CFG = cfg
        self.ready: bool = False

    def declare_unready(self, name):
        """
        Declare a parameter that cannot be set in the first (no db) phase of initialisation.

        :param name:
        :return:
        """
        if self.ready:
            raise ConfigException(f"Cannot declare {name} as unready on a ready object")
        self._unready_attributes.add(name)

    def get(self, key: str, default: Any = None) -> Any:
        """
        Expose attributes in a dictionary-like manner.

        :param key: The attribute name
        :param default: The default value to use if the attribute does not exist
        :return: The attribute value.
        """
        try:
            return getattr(self, key)
        except AttributeError:
            return default

    def __getattribute__(self, name: str) -> Any:
        """
        Throw an error if an unready attribute is accessed.

        :param name: attribute name
        :return: attribute value
        :raises: OWSConfigNotReady if entry is not fully initialised.
        """
        if name == "_unready_attributes":
            pass
        elif hasattr(self, "_unready_attributes") and name in self._unready_attributes:
            raise OWSConfigNotReady(f"The following parameters have not been initialised: {self._unready_attributes}")
        return object.__getattribute__(self, name)

    def __setattr__(self, name: str, val: Any) -> None:
        """
        Mark unready attributes off as they get set

        :param name: Attribute name
        :param val: Value to set the attribute to.
        """
        if name == "_unready_attributes":
            pass
        elif hasattr(self, "_unready_attributes") and name in self._unready_attributes:
            self._unready_attributes.remove(name)
        super().__setattr__(name, val)

    # Validate against database and prepare for use.
    def make_ready(self, *args, **kwargs) -> None:
        """
        Perform second phase initialisation with a database connection.

        In the base class we just confirm that all declared-unready attributes have been initialised,
        and set the ready flag.

        :param dc:
        :param args:
        :param kwargs:
        :return:
        """
        if self._unready_attributes:
            raise OWSConfigNotReady(f"The following parameters have not been initialised: {self._unready_attributes}")
        self.ready = True

#####################################################
# Metadata separation and translation.

# Label names for metadata separation and translation

FLD_TITLE = "title"
FLD_UNITS = "units"
FLD_ABSTRACT = "abstract"
FLD_KEYWORDS = "local_keywords"
FLD_FEES = "fees"
FLD_ACCESS_CONSTRAINTS = "access_constraints"
FLD_ATTRIBUTION = "attribution_title"
FLD_CONTACT_ORGANISATION = "contact_org"
FLD_CONTACT_POSITION = "contact_position"

class OWSMetadataConfig(OWSConfigEntry):
    """
    Config Entry abstract class that supports attributes that can be over-ridden with a message file or
    translations directory.
    """

    # Supported Metadata types.  Override them in the child classes as required
    # Each metadata type has special handling.

    METADATA_TITLE: bool = True
    METADATA_ABSTRACT: bool = True
    METADATA_KEYWORDS: bool = False
    METADATA_CONTACT_INFO: bool = False
    METADATA_FEES: bool = False
    METADATA_ACCESS_CONSTRAINTS: bool = False
    METADATA_ATTRIBUTION: bool = False
    METADATA_DEFAULT_BANDS: bool = False
    METADATA_VALUE_RULES: bool = False
    METADATA_LEGEND_UNITS: bool = False
    METADATA_TICK_LABELS: bool = False

    # Class registries, mapping metadata paths to their default value and whether the metadata value is
    # unique to that path, or has been inherited from a parent metadata path.
    _metadata_registry: dict[str, str] = {}
    _inheritance_registry: dict[str, bool] = {}

    _msg_src: Catalog | None = None

    # Inaccessible attributes to allow type checking
    abstract: str = ""
    attribution: Optional["datacube_ows.ows_configuration.AttributionCfg"] = None

    def get_obj_label(self) -> str:
        """Return the metadata path prefix for this object."""
        return "global"


    def can_inherit_from(self) -> Optional["OWSMetadataConfig"]:
        """
        The parent config object this object can inherit metadata from.

        :return: An instance of OWSMetadataConfig or None.
        """
        return None

    # Holders for managing inheritance.
    @property
    def default_title(self) -> str | None:
        return None

    @property
    def default_abstract(self) -> str | None:
        return None

    _keywords: set[str] = set()

    def parse_metadata(self, cfg: CFG_DICT) -> None:
        """
        Read some raw configuration for this object, and setup metadata handling.
        Must be called early in __init__() (before super().__init__().)

        :param cfg: The raw configuration for this object
        """
        # can_inherit_from() can be over-ridden by subclasses
        # pylint: disable=assignment-from-none
        inherit_from = self.can_inherit_from()
        if self.METADATA_TITLE:
            if self.default_title is not None:
                self.register_metadata(self.get_obj_label(), FLD_TITLE, cast(str, cfg.get(FLD_TITLE, self.default_title)))
            else:
                try:
                    self.register_metadata(self.get_obj_label(), FLD_TITLE, cast(str, cfg[FLD_TITLE]))
                except KeyError:
                    raise ConfigException(f"Entity {self.get_obj_label()} has no title.")
        if self.METADATA_ABSTRACT:
            local_abstract = cfg.get("abstract")
            if local_abstract is None and inherit_from is not None:
                self.register_metadata(self.get_obj_label(), FLD_ABSTRACT, inherit_from.abstract, inherited=True)
            elif local_abstract is None and self.default_abstract is not None:
                self.register_metadata(self.get_obj_label(), FLD_ABSTRACT, cast(str, self.default_abstract))
            elif local_abstract is None:
                raise ConfigException(f"Entity {self.get_obj_label()} has no abstract")
            else:
                self.register_metadata(self.get_obj_label(), "abstract", cast(str, local_abstract))
        if self.METADATA_KEYWORDS:
            local_keyword_set = set(cast(list[str], cfg.get("keywords", [])))
            self.register_metadata(self.get_obj_label(), FLD_KEYWORDS, ",".join(local_keyword_set))
            if inherit_from:
                keyword_set = inherit_from.keywords
            else:
                keyword_set = set()
            self._keywords = keyword_set.union(local_keyword_set)
        if self.METADATA_ATTRIBUTION:
            inheriting = False
            attrib = cast(dict[str, str], cfg.get("attribution"))
            if attrib is None and inherit_from is not None:
                attrib = inherit_from.attribution
                inheriting = True
            if attrib:
                attrib_title = attrib.get("title")
            else:
                attrib_title = None
            if attrib_title:
                self.register_metadata(self.get_obj_label(), FLD_ATTRIBUTION, attrib_title, inheriting)
        if self.METADATA_FEES:
            fees = cast(str, cfg.get(FLD_FEES))
            if not fees:
                fees = "none"
            self.register_metadata(self.get_obj_label(), FLD_FEES, fees)
        if self.METADATA_ACCESS_CONSTRAINTS:
            acc = cast(str, cfg.get("access_contraints"))
            if not acc:
                acc = "none"
            self.register_metadata(self.get_obj_label(), FLD_ACCESS_CONSTRAINTS, acc)
        if self.METADATA_CONTACT_INFO:
            cfg_contact_info: dict[str, str] = cast(dict[str, str], cfg.get("contact_info", {}))
            org = cfg_contact_info.get("organisation")
            position = cfg_contact_info.get("position")
            if org:
                self.register_metadata(self.get_obj_label(), FLD_CONTACT_ORGANISATION, org)
            if position:
                self.register_metadata(self.get_obj_label(), FLD_CONTACT_POSITION, position)
        if self.METADATA_DEFAULT_BANDS:
            band_map = cast(dict[str, list[str]], cfg)
            for k, v in band_map.items():
                if len(v):
                    self.register_metadata(self.get_obj_label(), k, v[0])
                else:
                    self.register_metadata(self.get_obj_label(), k, k)
        if self.METADATA_VALUE_RULES:
            # Note that parse_metadata must be called after the value map is parsed.
            for patch in self.patches:
                self.register_metadata(self.get_obj_label(), f"rule_{patch.idx}", patch.label)
        if self.METADATA_LEGEND_UNITS:
            units = cast(str, cfg.get(FLD_UNITS))
            if units is not None:
                self.register_metadata(self.get_obj_label(), FLD_UNITS, units)
        if self.METADATA_TICK_LABELS:
            # Note that parse_metadata must be called after legend ticks are parsed.
            for tick, lbl in zip(self.ticks, self.tick_labels):
                if any(c.isalpha() for c in lbl):
                    self.register_metadata(self.get_obj_label(), f"lbl_{tick}", lbl)

    @property
    def keywords(self) -> set[str]:
        """
        Return the keywords for this object (with inheritance, but without metadata separation or translation)
        :return: A set of keywords.
        """
        return self._keywords

    @classmethod
    def set_msg_src(cls, src: Catalog | None) -> None:
        """
        Allow all OWSMetadatConfig subclasses to share a common message catalog.
        :param src: A Message Catalog object
        """
        OWSMetadataConfig._msg_src = src

    def read_metadata(self, lbl: str, fld: str) -> str | None:
        """
        Read a general piece of metadata (potentially from another object).
        Resolution order:

        1) Via gettext translation if internationalisation is enabled.
        2) Via externalised message catalog, if configured.
        3) From raw config.

        :param lbl: Object label
        :param fld: Metadata type label
        :return: Displayable metadata.
        """
        lookup = ".".join([lbl, fld])
        if self.global_config().internationalised:
            trans = _(lookup)
            if trans != lookup:
                return trans
        if self._msg_src is not None:
            msg: Message | None = cast(Catalog, self._msg_src).get(lookup)
            if not msg:
                msg_: str | None = self._metadata_registry.get(lookup)
            else:
                msg_ = cast(str, msg.string)
            return msg_
        return self._metadata_registry.get(lookup)

    def read_inheritance(self, lbl: str, fld: str) -> bool:
        """
        Determine whether an arbitrary piece of metadata is inherited from a parent object.
        :param lbl: Object label
        :param fld: Metadata type label
        :return: True if the metadata is inherited from a parent object, False otherwise.
        """
        lookup = ".".join([lbl, fld])
        return self._inheritance_registry.get(lookup, False)

    def register_metadata(self, lbl: str, fld: str, val: str, inherited: bool = False) -> None:
        """
        Register a piece of metadata at config-parse time with it's raw config default value.

        :param lbl: Object label
        :param fld: Metadata type label
        :param val: The default value, as recorded in the raw config.
        :param inherited: If true, metadata is considered inherited and is not handled independently.
        """
        lookup = ".".join([lbl, fld])
        self._metadata_registry[lookup] = val
        self._inheritance_registry[lookup] = inherited

    def read_local_metadata(self, fld: str) -> str | None:
        """
        Read a general piece of metadata for this object.
        Resolution order:

        1) Via gettext translation if internationalisation is enabled.
        2) Via externalised message catalog, if configured.
        3) From raw config.

        :param fld: Metadata type label
        :return: Displayable metadata.
        """
        return self.read_metadata(self.get_obj_label(), fld)

    def is_inherited(self, fld: str) -> bool:
        """
        Determine whether an piece of metadata for this object is inherited from the parent object.
        :param fld: Metadata type label
        :return: True if the metadata is inherited from the parent object, False otherwise.
        """
        return self.read_inheritance(self.get_obj_label(), fld)

    def global_config(self) -> "datacube_ows.ows_configuration.OWSConfig":
        """
        Return the global config object.

        Should be over-ridden by child classes as appropriate.

        :return: The global (OWSConfig) configuration object.
        """
        return cast("datacube_ows.ows_configuration.OWSConfig", self)

    def __getattribute__(self, name: str) -> Any:
        """"
        Expose separated or internationalised metadata as attributes
        """
        if name in (FLD_TITLE, FLD_ABSTRACT, FLD_FEES,
                    FLD_ACCESS_CONSTRAINTS, FLD_CONTACT_POSITION, FLD_CONTACT_ORGANISATION,
                    FLD_UNITS):
            return self.read_local_metadata(name)
        elif name == FLD_KEYWORDS:
            kw = self.read_local_metadata(FLD_KEYWORDS)
            if kw:
                return set(kw.split(","))
            else:
                return set()
        elif name == FLD_ATTRIBUTION:
            return self.read_local_metadata(FLD_ATTRIBUTION)
        else:
            return super().__getattribute__(name)

###########################
# Inheritable configuration

class OWSEntryNotFound(ConfigException):
    """
    Exception thrown when looking up an indexed config entry that does not exist.
    """


class OWSIndexedConfigEntry(OWSConfigEntry):
    """
    A Config Entry object that can be looked up by name (i.e. so it can be inherited from)
    """
    INDEX_KEYS: list[str] = []

    def __init__(self, cfg: RAW_CFG, keyvals: dict[str, str], *args, **kwargs) -> None:
        """
        Validate and store keyvals for indexed lookup.

        :param cfg: Raw configuration for object.
        :param keyvals: Key values identifying this object
        :param args:
        :param kwargs:
        """
        super().__init__(cfg, *args, **kwargs)

        for k in self.INDEX_KEYS:
            if k not in keyvals:
                raise ConfigException(f"Key value {k} missing from keyvals: {keyvals!r}")
        self.keyvals = keyvals

    @classmethod
    def lookup_impl(cls, cfg: "datacube_ows.ows_configuration.OWSConfig",
                    keyvals: dict[str, str],
                    subs: CFG_DICT | None = None) -> "OWSIndexedConfigEntry":
        """
        Lookup a config entry of this type by identifying label(s)

        :param cfg:  The global config object that the desired object lives under.
        :param keyvals: Keyword dictionary of identifying label(s)
        :param subs:  Dictionary of keyword substitutions.  Used for e.g. looking up a style from a different layer.
        :return: The desired config object
        :raises: OWSEntryNotFound exception if no matching object found.
        """
        raise NotImplementedError()


# pylint: disable=abstract-method
class OWSExtensibleConfigEntry(OWSIndexedConfigEntry):
    """
    A configuration object that can inherit from and extend an existing configuration object of the same type.
    """
    def __init__(self,
                 cfg: RAW_CFG, keyvals: dict[str, str], global_cfg: "datacube_ows.ows_configuration.OWSConfig",
                 *args,
                 keyval_subs: dict[str, Any] | None = None,
                 keyval_defaults: dict[str, str] | None = None,
                 expanded: bool = False,
                 **kwargs) -> None:
        """
        Apply inheritance expansion and overrides to raw config.
        :param cfg: Raw (unexpanded) configuration
        :param keyvals: Dictionary of lookup keys to values
        :param global_cfg: global configuration object
        :param keyval_subs: (optional) Dictionary of keyword substitutions
        :param keyval_defaults: (optional) Dictionary of keyword defaults
        :param expanded: (optional, defaults to False) If true, assume expansion has already been applied.
        """
        if not expanded:
            cfg = self.expand_inherit(cast(CFG_DICT, cfg), global_cfg,
                                      keyval_subs=keyval_subs, keyval_defaults=keyval_defaults)

        super().__init__(cfg, keyvals, global_cfg=global_cfg, *args, **kwargs)

    @classmethod
    def expand_inherit(cls,
                       cfg: CFG_DICT, global_cfg: "datacube_ows.ows_configuration.OWSConfig",
                       keyval_subs: dict[str, Any] | None = None,
                       keyval_defaults: dict[str, str] | None = None) -> RAW_CFG:
        """
        Expand inherited config, and apply overrides.

        :param cfg: Unexpanded configuration
        :param global_cfg: Global config object
        :param keyval_subs: (optional) Dictionary of keyword substitutions
        :param keyval_defaults: (optional) Dictionary of keyword defaults
        :return: Fully expanded config, with inheritance and overrides applied.
        """
        if "inherits" in cfg:
            lookup = True
            # Precludes e.g. defaulting style lookup to current layer.
            lookup_keys: dict[str, str] = {}
            inherits = cast(dict[str, str], cfg["inherits"])
            for k in cls.INDEX_KEYS:
                if k not in inherits and keyval_defaults is not None and k not in keyval_defaults:
                    lookup = False
                    break
                if k in inherits:
                    lookup_keys[k] = inherits[k]
                elif keyval_defaults and k in keyval_defaults:
                    lookup_keys[k] = str(keyval_defaults[k])
            if lookup and lookup_keys:
                parent = cls.lookup_impl(global_cfg, keyvals=lookup_keys, subs=keyval_subs)
                # pylint: disable=protected-access
                parent_cfg = parent._raw_cfg
            else:
                parent_cfg = cfg["inherits"]
            cfg = deepinherit(cast(dict[str, Any], parent_cfg), cfg)
            cfg["inheritance_expanded"] = True
        return cfg

##################################
# Managing multiproduct flag-bands

class OWSFlagBandStandalone:
    """"
    Minimal proxy for OWSFlagBand, for use in StandAlone API which doesn't need anything more than the band name.
    """
    def __init__(self, band: str) -> None:
        self.pq_band = band
        self.canonical_band_name = band
        self.pq_names: list[str] = []
        self.pq_ignore_time = False
        self.pq_manual_merge = False
        self.pq_fuse_func: Optional[FunctionWrapper] = None


class OWSFlagBand(OWSConfigEntry):
    """
    Represents a flag band, which may come from the main product or a parallel secondary product.
    """
    def __init__(self, cfg: CFG_DICT, product_cfg: "datacube_ows.ows_configuration.OWSNamedLayer",
                 **kwargs) -> None:
        """
        Class constructor, first round initialisation

        :param cfg: Raw config
        :param product_cfg:  The OWSNamedLayer object this flag-band is associated with
        """
        super().__init__(cfg, **kwargs)
        cfg = cast(CFG_DICT, self._raw_cfg)
        self.layer = product_cfg
        pq_names = self.layer.parse_pq_names(cfg)
        self.pq_names = cast(list[str], pq_names["pq_names"])
        self.pq_low_res_names = pq_names["pq_low_res_names"]
        self.main_products = pq_names["main_products"]
        self.pq_band = str(cfg["band"])
        self.canonical_band_name = self.pq_band # Update for aliasing on make_ready
        if "fuse_func" in cfg:
            self.pq_fuse_func: Optional[FunctionWrapper] = FunctionWrapper(self.layer, cast(CFG_DICT, cfg["fuse_func"]))
        else:
            self.pq_fuse_func = None
        self.pq_ignore_time = bool(cfg.get("ignore_time", False))
        self.ignore_info_flags = cast(list[str], cfg.get("ignore_info_flags", []))
        self.pq_manual_merge = bool(cfg.get("manual_merge", False))
        self.declare_unready("pq_products")
        self.declare_unready("flags_def")
        self.declare_unready("info_mask")

    # pylint: disable=attribute-defined-outside-init
    def make_ready(self, *args: Any, **kwargs: Any) -> None:
        """
        Second round (db-aware) intialisation.

        :param dc: A Datacube object
        """
        # pyre-ignore[16]
        self.pq_products: list[Product] = []
        # pyre-ignore[16]
        self.pq_low_res_products: list[Product] = []
        for pqn in self.pq_names:
            if pqn is not None:
                pq_product = self.layer.dc.index.products.get_by_name(pqn)
                if pq_product is None:
                    raise ConfigException(f"Could not find flags product {pqn} for layer {self.layer.name} in datacube")
                self.pq_products.append(pq_product)
        for pqn in self.pq_low_res_names:
            if pqn is not None:
                pq_product = self.layer.dc.index.products.get_by_name(pqn)
                if pq_product is None:
                    raise ConfigException(f"Could not find flags low_res product {pqn} for layer {self.layer.name} in datacube")
                self.pq_low_res_products.append(pq_product)

        # Resolve band alias if necessary.
        if self.main_products:
            try:
                self.canonical_band_name = self.layer.band_idx.band(self.pq_band)
            except ConfigException:
                pass

    # pyre-ignore[16]
        self.info_mask: int = ~0
        # A (hopefully) representative product
        product = self.pq_products[0]
        try:
            meas = product.lookup_measurements([str(self.canonical_band_name)])[str(self.canonical_band_name)]
        except KeyError:
            raise ConfigException(
                f"Band {self.pq_band} does not exist in product {product.name} - cannot be used as a flag band for layer {self.layer.name}.")
        if "flags_definition" not in meas:
            raise ConfigException(f"Band {self.pq_band} in product {product.name} has no flags_definition in ODC - cannot be used as a flag band for layer {self.layer.name}.")
        # pyre-ignore[16]
        self.flags_def: dict[str, dict[str, RAW_CFG]] = meas["flags_definition"]
        for bitname in self.ignore_info_flags:
            bit = self.flags_def[bitname]["bits"]
            if not isinstance(bit, int):
                continue
            flag = 1 << bit
            self.info_mask &= ~flag
        super().make_ready(*args, **kwargs)


FlagBand = OWSFlagBand | OWSFlagBandStandalone


class FlagProductBands(OWSConfigEntry):
    """
    A collection of flag bands for a layer that all come from the same product (or multi-product product collection).
    """
    def __init__(self, flag_band: FlagBand,
                 layer: "datacube_ows.ows_configuration.OWSNamedLayer") -> None:
        """
        Class constructor, first round initialisation

        :param flag_band: A Flag-band object
        :param layer: the named layer these flag-bands are associated with.
        """
        super().__init__({})
        self.layer = layer
        self.bands: set[str] = set()
        self.bands.add(str(flag_band.canonical_band_name))
        self.flag_bands = {flag_band.pq_band: flag_band}
        self.product_names = tuple(flag_band.pq_names)
        self.ignore_time = flag_band.pq_ignore_time
        self.declare_unready("products")
        self.declare_unready("low_res_products")
        self.manual_merge = bool(flag_band.pq_manual_merge)
        self.fuse_func = flag_band.pq_fuse_func
        # pyre-ignore[16]
        self.main_product = self.products_match(layer.product_names)

    def products_match(self, product_names: Iterable[str]) -> bool:
        """
        Compare a list of product names to this objects list of product names.

        :param product_names: The product names to compare to this object
        :return: True if the product names lists match, False otherwise.
        """
        return tuple(product_names) == self.product_names

    def add_flag_band(self, fb: FlagBand) -> None:
        """
        Add an additional flag band to this collection

        :param fb:  A flag-band object.  Must match product_names with this collection.
        """
        self.flag_bands[fb.pq_band] = fb
        self.bands.add(fb.pq_band)
        if fb.pq_manual_merge:
            fb.pq_manual_merge = True
        if fb.pq_fuse_func and self.fuse_func and fb.pq_fuse_func != self.fuse_func:
            raise ConfigException(f"Fuse functions for flag bands in product set {self.product_names} do not match")
        if fb.pq_ignore_time != self.ignore_time:
            raise ConfigException(f"ignore_time option for flag bands in product set {self.product_names} do not match")
        elif fb.pq_fuse_func and not self.fuse_func:
            self.fuse_func = fb.pq_fuse_func
        self.declare_unready("products")
        self.declare_unready("low_res_products")

    # pylint: disable=attribute-defined-outside-init
    def make_ready(self, *args, **kwargs) -> None:
        """
        Second round (db-aware) intialisation.

        :param dc: A Datacube object
        """
        for fb in self.flag_bands.values():
            fb = cast(OWSFlagBand, fb)
            # pyre-ignore [16]
            self.products: list[Product] = fb.pq_products
            # pyre-ignore [16]
            self.low_res_products: list[Product] = fb.pq_low_res_products
            break
        if self.main_product:
            self.bands = set(self.layer.band_idx.band(b) for b in self.bands)
        super().make_ready(*args, **kwargs)

    @classmethod
    def build_list_from_masks(cls, masks: Iterable["datacube_ows.styles.base.StyleMask"],
                              layer: "datacube_ows.ows_configuration.OWSNamedLayer") -> list["FlagProductBands"]:
        """
        Class method to instantiate a list of FlagProductBands from a list of style masks.

        Bands from the same product (or multi-product collection) will be grouped into the same FlagProductBand

        :param masks: A list of StyleMask objects.
        :param layer: A named layer object
        :return: A list of FlagProductBands objects
        """
        flag_products: list["FlagProductBands"] = []
        for mask in masks:
            handled = False
            for fp in flag_products:
                if fp.products_match(mask.flag_band.pq_names):
                    fp.add_flag_band(mask.flag_band)
                    handled = True
                    break
            if not handled:
                flag_products.append(cls(mask.flag_band, layer))
        return flag_products

    @classmethod
    def build_list_from_flagbands(cls, flagbands: Iterable[OWSFlagBand],
                                  layer: "datacube_ows.ows_configuration.OWSNamedLayer") -> list["FlagProductBands"]:
        """
        Class method to instantiate a list of FlagProductBands from a list of OWS Flag Bands.

        Bands from the same product (or multi-product collection) will be grouped into the same FlagProductBand

        :param masks: A list of OWSFlagBand objects.
        :param layer: A named layer object
        :return: A list of FlagProductBands objects
        """
        flag_products: list["FlagProductBands"] = []
        for fb in flagbands:
            handled = False
            for fp in flag_products:
                if fp.products_match(fb.pq_names):
                    fp.add_flag_band(fb)
                handled = True
                break
            if not handled:
                flag_products.append(cls(fb, layer))
        return flag_products


FlagSpec = dict[str, bool | str]


class AbstractMaskRule(OWSConfigEntry):
    def __init__(self, band: str, cfg: CFG_DICT, mapper: Callable[[str], str] = lambda x: x) -> None:
        super().__init__(cfg)
        self.band = mapper(band)
        self.parse_rule_spec(cfg)

    @property
    def context(self) -> str:
        return "a mask rule"

    VALUES_LABEL = "values"
    def parse_rule_spec(self, cfg: CFG_DICT) -> None:
        self.flags: list[FlagSpec] | FlagSpec | None = None
        self.or_flags: bool | list[bool] = False
        self.values: list[list[int]] | list[int] | None = None
        self.invert: bool | list[bool] = bool(cfg.get("invert", False))
        if "flags" in cfg:
            flags = cast(FlagSpec, cfg["flags"])
            if "or" in flags and "and" in flags:
                raise ConfigException(
                    f"ValueMap rule in {self.context} combines 'and' and 'or' rules")
            elif "or" in flags:
                self.or_flags = True
                flags = cast(FlagSpec, flags["or"])
            elif "and" in flags:
                flags = cast(FlagSpec, flags["and"])
            self.flags = flags
        else:
            self.flags = None

        if "values" in cfg:
            val: Any = cfg["values"]
        else:
            val = None
        if val is None:
            self.values = None
        else:
            if isinstance(val, int):
                self.values = [cast(int, val)]
            else:
                self.values = cast(list[int], val)

        if not self.flags and not self.values:
            raise ConfigException(
                f"Mask rule in {self.context} must have a non-empty 'flags' or 'values' section.")
        if self.flags and self.values:
            raise ConfigException(
                f"Mask rule in {self.context} has both a 'flags' and a 'values' section - choose one.")

    def create_mask(self, data: DataArray) -> DataArray | None:
        """
        Create a mask from raw flag band data.

        :param data: Raw flag data, assumed to be for this rule's flag band.
        :return: A boolean DataArray, True where the data matches this rule
        """
        if self.values:
            mask: DataArray | None = None
            for v in cast(list[int], self.values):
                vmask = data == v
                if mask is None:
                    mask = vmask
                else:
                    mask |= vmask
        elif self.or_flags:
            mask = None
            for f in cast(CFG_DICT, self.flags).items():
                d = {f[0]: f[1]}
                if mask is None:
                    mask = make_mask(data, **d)
                else:
                    mask |= make_mask(data, **d)
        else:
            mask = make_mask(data, **cast(CFG_DICT, self.flags))
        if mask is not None and self.invert:
            mask = ~mask # pylint: disable=invalid-unary-operand-type
        return mask


# Function wrapper for configurable functional elements
class FunctionWrapper:
    """
    Function wrapper for configurable functional elements
    """

    def __init__(self,
                 product_or_style_cfg: OWSExtensibleConfigEntry | None,
                 func_cfg: str | CFG_DICT | F,
                 stand_alone: bool = False) -> None:
        """

        :param product_or_style_cfg: An instance of either NamedLayer or Style,
                the context in which the wrapper operates.
        :param func_cfg: A function or a configuration dictionary representing a function.
        :param stand_alone: Optional boolean.
                If False (the default) then only configuration dictionaries will be accepted.
        """
        self.style_or_layer_cfg = product_or_style_cfg
        if callable(func_cfg):
            if not stand_alone:
                raise ConfigException(
                    "Directly including callable objects in configuration is no longer supported. Please reference callables by fully qualified name.")
            self._func: Callable = func_cfg
            self._args: list[RAW_CFG] = []
            self._kwargs: CFG_DICT = {}
            self.band_mapper: Callable[[str], str] | None = None
            self.pass_layer_cfg = False
        elif isinstance(func_cfg, str):
            self._func = get_function(func_cfg)
            self._args = []
            self._kwargs = {}
            self.band_mapper = None
            self.pass_layer_cfg = False
        else:
            if stand_alone and callable(func_cfg["function"]):
                self._func = func_cfg["function"]
            elif callable(func_cfg["function"]):
                raise ConfigException(
                    "Directly including callable objects in configuration is no longer supported. Please reference callables by fully qualified name.")
            else:
                self._func = get_function(cast(str, func_cfg["function"]))
            self._args = cast(list[RAW_CFG], func_cfg.get("args", []))
            self._kwargs = cast(CFG_DICT, func_cfg.get("kwargs", {})).copy()
            self.pass_layer_cfg = bool(func_cfg.get("pass_layer_cfg", False))
            if "pass_product_cfg" in func_cfg:
                _LOG.warning("WARNING: pass_product_cfg in function wrapper definitions has been renamed "
                             "'mapped_bands'.  Please update your config accordingly")
            if func_cfg.get("mapped_bands", func_cfg.get("pass_product_cfg", False)):
                if hasattr(product_or_style_cfg, "band_idx"):
                    # NamedLayer
                    from datacube_ows.ows_configuration import OWSNamedLayer
                    named_layer = cast(OWSNamedLayer, product_or_style_cfg)
                    b_idx = named_layer.band_idx
                    self.band_mapper = b_idx.band
                else:
                    # Style
                    from datacube_ows.styles import StyleDef
                    style = cast(StyleDef, product_or_style_cfg)
                    b_idx = style.product.band_idx
                    delocaliser = style.local_band
                    self.band_mapper = lambda b: b_idx.band(delocaliser(b))
            else:
                self.band_mapper = None

    def __call__(self, *args, **kwargs) -> Any:
        if args and self._args:
            calling_args: Iterable[Any] = chain(args, self._args)
        elif args:
            calling_args = args
        else:
            calling_args = self._args
        if kwargs and self._kwargs:
            calling_kwargs: dict[str, Any] = self._kwargs.copy()
            calling_kwargs.update(kwargs)
        elif kwargs:
            calling_kwargs = kwargs.copy()
        else:
            calling_kwargs = self._kwargs.copy()

        if self.band_mapper:
            calling_kwargs["band_mapper"] = self.band_mapper

        if self.pass_layer_cfg:
            calling_kwargs['layer_cfg'] = self.style_or_layer_cfg

        return self._func(*calling_args, **calling_kwargs)


def get_function(func: F | str) -> F:
    """Converts a config entry to a function, if necessary

    :param func: Either a Callable object or a fully qualified function name str, or None
    :return: a Callable object, or None
    """
    if func is not None and not callable(func):
        mod_name, func_name = func.rsplit('.', 1)
        try:
            mod = import_module(mod_name)
            func = getattr(mod, func_name)
        except (ImportError, ModuleNotFoundError, ValueError, AttributeError):
            raise ConfigException(f"Could not import python object: {func}")
        assert callable(func)
    return cast(F, func)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import importlib
import click
import psycopg2
import re
import sqlalchemy

from datacube import Datacube
from datacube_ows.index import AbortRun


def get_sqlconn(dc: Datacube) -> sqlalchemy.Connection:
    """
    Extracts a SQLAlchemy database connection from a Datacube object.

    :param dc: An initialised Datacube object
    :return: A SQLAlchemy database connection object.
    """
    # pylint: disable=protected-access
    return dc.index._db._engine.connect()  # type: ignore[attr-defined]


def run_sql(dc: Datacube, driver_name: str, path: str, **params: str) -> bool:
    if not importlib.resources.files("datacube_ows").joinpath(f"sql/{driver_name}/{path}").is_dir():
        print("Cannot find SQL resource directory - check your datacube-ows installation")
        return False

    files = sorted(
        importlib.resources.files("datacube_ows").joinpath(f"sql/{driver_name}/{path}").iterdir()  # type: ignore[type-var]
    )

    filename_req_pattern = re.compile(r"\d+[_a-zA-Z0-9]+_requires_(?P<reqs>[_a-zA-Z0-9]+)\.sql")
    filename_pattern = re.compile(r"\d+[_a-zA-Z0-9]+\.sql")
    conn = get_sqlconn(dc)
    all_ok: bool = True
    for fi in files:
        f = fi.name
        match = filename_pattern.fullmatch(f)
        if not match:
            click.echo(f"Illegal SQL filename: {f} (skipping)")
            all_ok = False
            continue
        req_match = filename_req_pattern.fullmatch(f)
        if req_match:
            reqs = req_match.group("reqs").split("_")
        else:
            reqs = []
        if reqs:
            try:
                kwargs = {v: params[v] for v in reqs}
            except KeyError as e:
                click.echo(f"Required parameter {e} for file {f} not supplied - skipping")
                all_ok = False
                continue
        else:
            kwargs = {}
        ref = importlib.resources.files("datacube_ows").joinpath(f"sql/{driver_name}/{path}/{f}")
        with ref.open("rb") as fp:
            sql = ""
            first = True
            for line in fp:
                sline = str(line, "utf-8")
                if first and sline.startswith("--"):
                    if reqs:
                        click.echo(f" - Running {sline[2:].format(**kwargs)}")
                    else:
                        click.echo(f" - Running {sline[2:]}")
                else:
                    sql = sql + "\n" + sline
                first = False
        if reqs:
            sql = sql.format(**kwargs)
        try:
            result = conn.execute(sqlalchemy.text(sql))
            click.echo(f"    ...  succeeded(?) with rowcount {result.rowcount}")

        except sqlalchemy.exc.ProgrammingError as e:
            if isinstance(e.orig, psycopg2.errors.InsufficientPrivilege):
                click.echo(
                    f"Insufficient Privileges (user {dc.index.environment.db_username}). Schema altering actions should be run by a role with admin privileges"
                )
                raise AbortRun() from None
            elif isinstance(e.orig, psycopg2.errors.DuplicateObject):
                if f.endswith('_ignore_duplicates.sql'):
                    click.echo(f"Ignoring 'already exists' error")
                else:
                    raise e from None
            else:
                raise e from e
    return all_ok
    conn.close()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import click
import datetime

from threading import Lock
from typing import Any, Iterable, Type
from uuid import UUID

from odc.geo import Geometry, CRS
from datacube import Datacube
from datacube.model import Product, Dataset, Range

from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.index.api import OWSAbstractIndex, OWSAbstractIndexDriver, LayerSignature, LayerExtent, TimeSearchTerm
from datacube_ows.index.sql import run_sql
from .product_ranges import create_range_entry as create_range_entry_impl, get_ranges as get_ranges_impl
from ...utils import default_to_utc


class OWSPostgisIndex(OWSAbstractIndex):
    name: str = "postgis"

    # method to delete obsolete schemas etc.
    def cleanup_schema(self, dc: Datacube):
        # No obsolete schema for postgis databases to clean up.
        pass

    # Schema creation method
    def create_schema(self, dc: Datacube):
        click.echo("Creating/updating schema and tables...")
        self._run_sql(dc, "ows_schema/create")

    # Permission management method
    def grant_perms(self, dc: Datacube, role: str, read_only: bool = False):
        if read_only:
            self._run_sql(dc, "ows_schema/grants/read_only", role=role)
        else:
            self._run_sql(dc, "ows_schema/grants/read_write", role=role)

    # Spatiotemporal index update method (e.g. refresh materialised views)
    def update_geotemporal_index(self, dc: Datacube):
        # Native ODC geotemporal index used in postgis driver.
        pass

    def create_range_entry(self, layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]) -> None:
        create_range_entry_impl(layer, cache)

    def get_ranges(self, layer: OWSNamedLayer) -> LayerExtent | None:
        return get_ranges_impl(layer)

    def _query(self,
               layer: OWSNamedLayer,
               times: Iterable[TimeSearchTerm] | None = None,
               geom: Geometry | None = None,
               products: Iterable[Product] | None = None
               ) -> dict[str, Any]:
        query: dict[str, Any] = {}
        if geom:
            query["geopolygon"] = self._prep_geom(layer, geom)
        if products is not None:
            query["product"] = [p.name for p in products]
        if times is not None:

            def normalise_to_dtr(unnorm: datetime.datetime | datetime.date) -> tuple[datetime.datetime, datetime.datetime]:
                if isinstance(unnorm, datetime.datetime):
                    st: datetime.datetime = default_to_utc(unnorm)
                    tmax = st + datetime.timedelta(seconds=1)
                elif isinstance(t, datetime.date):
                    st = datetime.datetime(unnorm.year, unnorm.month, unnorm.day, tzinfo=datetime.timezone.utc)
                    tmax = st + datetime.timedelta(days=1)
                else:
                    raise ValueError("Not a datetime object")
                return st, tmax

            time_args = []
            for t in times:
                if isinstance(t, (datetime.date, datetime.datetime)):
                    start, tmax = normalise_to_dtr(t)
                    time_args.append(Range(start, tmax))
                else:
                    st, et = t
                    st, _ = normalise_to_dtr(st)
                    et, _ = normalise_to_dtr(et)
                    time_args.append(Range(st, et))
            if len(time_args) > 1:
                raise ValueError("Huh?")
            query["time"] = time_args[0]
        return query

    def ds_search(self,
                  layer: OWSNamedLayer,
                  times: Iterable[TimeSearchTerm] | None = None,
                  geom: Geometry | None = None,
                  products: Iterable[Product] | None = None
                  ) -> Iterable[Dataset]:
        return layer.dc.index.datasets.search(**self._query(layer, times, geom, products))

    def dsid_search(self,
                    layer: OWSNamedLayer,
                    times: Iterable[TimeSearchTerm] | None = None,
                    geom: Geometry | None = None,
                    products: Iterable[Product] | None = None
                    ) -> Iterable[UUID]:
        for ds in layer.dc.index.datasets.search_returning(field_names=["id"],
                                                           **self._query(layer, times, geom, products)):
            yield ds.id  # type: ignore[attr-defined]

    def count(self,
              layer: OWSNamedLayer,
              times: Iterable[TimeSearchTerm] | None = None,
              geom: Geometry | None = None,
              products: Iterable[Product] | None = None
              ) -> int:
        return layer.dc.index.datasets.count(**self._query(layer, times, geom, products))

    def extent(self,
               layer: OWSNamedLayer,
               times: Iterable[TimeSearchTerm] | None = None,
               geom: Geometry | None = None,
               products: Iterable[Product] | None = None,
               crs: CRS | None = None
               ) -> Geometry | None:
        if crs is None:
            crs = CRS("epsg:4326")
        return layer.dc.index.datasets.spatial_extent(
            self.dsid_search(layer, times=times, geom=geom, products=products),
            crs=crs
        )

    def _run_sql(self, dc: Datacube, path: str, **params: str) -> bool:
        return run_sql(dc, self.name, path, **params)


pgisdriverlock = Lock()


class OWSPostgisIndexDriver(OWSAbstractIndexDriver):
    _driver = None
    @classmethod
    def ows_index_class(cls) -> Type[OWSAbstractIndex]:
        return OWSPostgisIndex

    @classmethod
    def ows_index(cls) -> OWSAbstractIndex:
        with pgisdriverlock:
            if cls._driver is None:
                cls._driver = OWSPostgisIndex()
        return cls._driver


def ows_index_driver_init():
    return OWSPostgisIndexDriver()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import re
from datetime import datetime
from itertools import chain
from typing import cast

import numpy
import xarray
from datacube.model import Dataset
from datacube.utils.masking import mask_to_dict
from odc.geo import geom
from odc.geo.geobox import GeoBox
from pandas import Timestamp

from datacube_ows.config_utils import CFG_DICT, RAW_CFG, ConfigException
from datacube_ows.http_utils import (FlaskResponse, html_json_response,
                                     json_response)
from datacube_ows.loading import DataStacker, ProductBandQuery
from datacube_ows.ows_configuration import OWSNamedLayer, get_config
from datacube_ows.styles import StyleDef
from datacube_ows.time_utils import dataset_center_time, tz_for_geometry
from datacube_ows.utils import log_call
from datacube_ows.wms_utils import (GetFeatureInfoParameters,
                                    img_coords_to_geopoint)


@log_call
def get_s3_browser_uris(datasets: dict[ProductBandQuery, xarray.DataArray],
                        pt: geom.Geometry | None = None,
                        s3url: str = "", s3bucket: str = "") -> set[str]:
    uris = []
    last_crs = None
    for pbq, dss in datasets.items():
        if pbq.main:
            for tds in dss:
                for ds in tds.values.item():
                    if pt and ds.extent:
                        if ds.crs != last_crs:
                            pt_native = pt.to_crs(ds.crs)
                            last_crs = ds.crs
                        else:
                            pt_native = pt
                        if ds.extent.contains(pt_native):
                            uris.append(ds.uris)
                    else:
                        uris.append(ds.uris)
            break

    uris = list(chain.from_iterable(uris))
    unique_uris = set(uris)

    regex = re.compile(r"s3:\/\/(?P<bucket>[a-zA-Z0-9_\-\.]+)\/(?P<prefix>[\S]+)/[a-zA-Z0-9_\-\.]+.(yaml|json)")

    # convert to browsable link
    def convert(uri: str) -> str:
        uri_format = "http://{bucket}.s3-website-ap-southeast-2.amazonaws.com/?prefix={prefix}"
        uri_format_prod = str(s3url) + "/?prefix={prefix}"
        result = regex.match(uri)
        if result is not None:
            if result.group("bucket") == str(s3bucket):
                new_uri = uri_format_prod.format(prefix=result.group("prefix"))
            else:
                new_uri = uri_format.format(bucket=result.group("bucket"),
                                            prefix=result.group("prefix"))
        else:
            new_uri = uri
        return new_uri

    formatted = {convert(uri) for uri in unique_uris}

    return formatted


@log_call
def _make_band_dict(prod_cfg: OWSNamedLayer, pixel_dataset: xarray.Dataset) -> dict[str, dict[str, bool | str] | str]:
    band_dict: dict[str, dict[str, bool | str] | str] = {}
    for k, v in pixel_dataset.data_vars.items():
        band_val = pixel_dataset[k].item()
        flag_def = pixel_dataset[k].attrs.get("flags_definition")
        if flag_def:
            try:
                flag_dict = mask_to_dict(flag_def, band_val)
            except TypeError as te:
                logging.warning('Working around for float bands')
                flag_dict = mask_to_dict(flag_def, int(band_val))
            ret_val: dict[str, bool | str] = {}
            for flag, val in flag_dict.items():
                if not val:
                    continue
                if val == True:
                    ret_val[flag_def[flag].get('description', flag)] = True
                else:
                    ret_val[flag_def[flag].get('description', flag)] = val
            band_dict[k] = ret_val
        else:
            try:
                band_lbl = prod_cfg.band_idx.band_label(k)
                assert k is not None   # for type checker
                if band_val == pixel_dataset[k].nodata or numpy.isnan(band_val):
                    band_dict[band_lbl] = "n/a"
                else:
                    band_dict[band_lbl] = band_val
            except ConfigException:
                pass
    return band_dict


@log_call
def _make_derived_band_dict(pixel_dataset: xarray.Dataset, style_index: dict[str, StyleDef]) -> dict[str, int | float]:
    """Creates a dict of values for bands derived by styles.
    This only works for styles with an `index_function` defined.

    :param xarray.Dataset pixel_dataset: A 1x1 pixel dataset containing band arrays
    :param dict(str, StyleCfg) style_index: dict of style configuration dicts
    :return: dict of style names to derived value
    """
    derived_band_dict = {}
    for style_name, style in style_index.items():
        if not style.include_in_feature_info:
            continue

        if any(pixel_dataset[band] == pixel_dataset[band].nodata for band in style.needed_bands):
            continue

        value = style.index_function(pixel_dataset).item()
        derived_band_dict[style_name] = value if not numpy.isnan(value) else "n/a"
    return derived_band_dict


def geobox_is_point(geobox: GeoBox) -> bool:
    return geobox.height == 1 and geobox.width == 1


@log_call
def feature_info(args: dict[str, str]) -> FlaskResponse:
    # pylint: disable=too-many-nested-blocks, too-many-branches, too-many-statements, too-many-locals
    # Parse GET parameters
    params = GetFeatureInfoParameters(args)
    feature_json: CFG_DICT = {}

    geo_point = img_coords_to_geopoint(params.geobox, params.i, params.j)
    # shrink geobox to point
    # Prepare to extract feature info
    if geobox_is_point(params.geobox):
        # request geobox is already 1x1
        geo_point_geobox = params.geobox
    else:
        # Make a 1x1 pixel geobox
        geo_point_geobox = GeoBox.from_geopolygon(
            geo_point, params.geobox.resolution, crs=params.geobox.crs)
    tz = tz_for_geometry(geo_point_geobox.geographic_extent)
    stacker = DataStacker(params.layer, geo_point_geobox, params.times)
    # --- Begin code section requiring datacube.
    cfg = get_config()
    all_time_datasets = stacker.datasets_all_time(point=geo_point)

    # Taking the data as a single point so our indexes into the data should be 0,0
    h_coord = cast(str, cfg.published_CRSs[params.crsid]["horizontal_coord"])
    v_coord = cast(str, cfg.published_CRSs[params.crsid]["vertical_coord"])
    s3_bucket = cfg.s3_bucket
    s3_url = cfg.s3_url
    isel_kwargs = {
        h_coord: 0,
        v_coord: 0
    }
    if any(all_time_datasets):
        # Group datasets by time, load only datasets that match the idx_date
        global_info_written = False
        feature_json["data"] = []
        fi_date_index: dict[datetime, RAW_CFG] = {}
        time_datasets = stacker.datasets(all_flag_bands=True, point=geo_point)
        data = stacker.data(time_datasets, skip_corrections=True)
        if data is not None:
            for dt in data.time.values:
                td = data.sel(time=dt)
                # Global data that should apply to all dates, but needs some data to extract
                if not global_info_written:
                    global_info_written = True
                    # Non-geographic coordinate systems need to be projected onto a geographic
                    # coordinate system.  Why not use EPSG:4326?
                    # Extract coordinates in CRS
                    data_x = getattr(td, h_coord)
                    data_y = getattr(td, v_coord)

                    x = data_x[isel_kwargs[h_coord]].item()
                    y = data_y[isel_kwargs[v_coord]].item()
                    pt = geom.point(x, y, params.crs)

                    # Project to EPSG:4326
                    crs_geo = geom.CRS("EPSG:4326")
                    ptg = pt.to_crs(crs_geo)

                    # Capture lat/long coordinates
                    feature_json["lon"], feature_json["lat"] = ptg.coords[0]

                date_info: CFG_DICT = {}

                ds: Dataset | None = None
                for pbq, dss in time_datasets.items():
                    if pbq.main:
                        ds = dss.sel(time=dt).values.tolist()[0]
                        break
                assert ds is not None
                if params.layer.multi_product:
                    if "platform" in ds.metadata_doc:
                        date_info["source_product"] = "%s (%s)" % (ds.type.name, ds.metadata_doc["platform"]["code"])
                    else:
                        date_info["source_product"] = ds.type.name

                # Extract data pixel
                pixel_ds: xarray.Dataset = td.isel(**isel_kwargs)  # type: ignore[arg-type]

                # Get accurate timestamp from dataset
                assert ds.time is not None  # For type checker
                if params.layer.time_resolution.is_summary():
                    date_info["time"] = ds.time.begin.strftime("%Y-%m-%d")
                else:
                    date_info["time"] = dataset_center_time(ds).strftime("%Y-%m-%d %H:%M:%S %Z")
                # Collect raw band values for pixel and derived bands from styles
                date_info["bands"] = cast(RAW_CFG, _make_band_dict(params.layer, pixel_ds))
                derived_band_dict = cast(RAW_CFG, _make_derived_band_dict(pixel_ds, params.layer.style_index))
                if derived_band_dict:
                    date_info["band_derived"] = derived_band_dict
                # Add any legacy custom-defined fields.
                for k, f in params.layer.legacy_feature_info_custom_includes.items():
                    # legacy function signature: pass in band and index data as a dictionary
                    date_info[k] = f(date_info["bands"])

                # Any custom-defined layer fields
                # (entries from the legacy are overwritten by the new if both exist)
                for k, f in params.layer.feature_info_custom_includes.items():
                    # New function signature: pass in:
                    # * a single pixel (1x1x1) multiband xarray Dataset, and
                    # * the ODC Dataset model (i.e. full ODC metadata)
                    date_info[k] = f(pixel_ds, ds)

                if params.style is not None:
                    # Any custom-defined style fields
                    # (style definitions override layer-level entries where both exist)
                    for k, f in params.style.feature_info_includes.items():
                        # Function signature: pass in:
                        # * a single pixel (1x1x1) multiband xarray Dataset, and
                        # * the ODC Dataset model (i.e. full ODC metadata)
                        date_info[k] = f(pixel_ds, ds)

                cast(list[RAW_CFG], feature_json["data"]).append(date_info)
                fi_date_index[dt] = cast(dict[str, list[RAW_CFG]], feature_json)["data"][-1]
            # Multidate custom includes reflect all selected times on multidate requests,
            # and are added as an extra all-time data record after the date ones.
            times = list(data.time.values)
            if len(times) > 1 and params.style is not None:
                mdh = params.style.get_multi_date_handler(times)
                if mdh is not None:
                    date_info = {
                        "time": "all"
                    }
                    for k, f in mdh.feature_info_includes.items():
                        # Function signature: pass in:
                        # * a multi-date single pixel (1x1xn) multiband xarray Dataset,
                        date_info[k] = f(data)
                cast(list[RAW_CFG], feature_json["data"]).append(date_info)
        feature_json["data_available_for_dates"] = []
        pt_native = None
        for d in all_time_datasets.coords["time"].values:
            dt_datasets = all_time_datasets.sel(time=d)
            for ds in dt_datasets.values.item():
                assert ds is not None  # For type checker
                if pt_native is None:
                    pt_native = geo_point.to_crs(ds.crs)
                elif pt_native.crs != ds.crs:
                    pt_native = geo_point.to_crs(ds.crs)
                if ds.extent and ds.extent.contains(pt_native):
                    # tolist() converts a numpy datetime64 to a python datatime
                    dt = Timestamp(stacker.group_by.group_by_func(ds)).to_pydatetime()
                    if params.layer.time_resolution.is_subday():
                        cast(list[RAW_CFG], feature_json["data_available_for_dates"]).append(dt.isoformat())
                    else:
                        cast(list[RAW_CFG], feature_json["data_available_for_dates"]).append(dt.strftime("%Y-%m-%d"))
                    break
        if time_datasets:
            feature_json["data_links"] = cast(
                RAW_CFG,
                sorted(get_s3_browser_uris(time_datasets, pt_native, s3_url, s3_bucket)))
        else:
            feature_json["data_links"] = []
        if params.layer.feature_info_include_utc_dates:
            unsorted_dates: list[str] = []
            for tds in all_time_datasets:
                for ds in tds.values.item():
                    assert ds is not None and ds.time is not None  # for type checker
                    if params.layer.time_resolution.is_solar():
                        unsorted_dates.append(ds.center_time.strftime("%Y-%m-%d"))
                    elif params.layer.time_resolution.is_subday():
                        unsorted_dates.append(ds.time.begin.isoformat())
                    else:
                        unsorted_dates.append(ds.time.begin.strftime("%Y-%m-%d"))
            feature_json["data_available_for_utc_dates"] = sorted(
                d.center_time.strftime("%Y-%m-%d") for d in all_time_datasets)

    result: CFG_DICT = {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "properties": feature_json,
                "geometry": {
                    "type": "Point",
                    "coordinates": geo_point.coords[0]
                }
            }
        ]
    }
    if params.format == "text/html":
        return html_json_response(result, cfg)
    else:
        return json_response(result, cfg)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import collections
import logging

from dateutil.parser import parse
from odc.geo.geobox import GeoBox
from ows.wcs.v20 import ScaleAxis, ScaleExtent, ScaleSize, Slice, Trim
from rasterio import MemoryFile

from datacube_ows.loading import DataStacker
from datacube_ows.ogc_exceptions import WCS2Exception
from datacube_ows.ows_configuration import get_config
from datacube_ows.resource_limits import ResourceLimited
from datacube_ows.utils import default_to_utc
from datacube_ows.wcs_scaler import WCSScaler, WCSScalerUnknownDimension

# from datacube_ows.wcs_utils import get_bands_from_styles

_LOG = logging.getLogger(__name__)


def uniform_crs(cfg, crs):
    " Helper function to transform a URL style EPSG definition to an 'EPSG:nnn' one "
    if crs.startswith('http://www.opengis.net/def/crs/EPSG/'):
        code = crs.rpartition('/')[-1]
        crs = 'EPSG:%s' % code
    elif crs.startswith('urn:ogc:def:crs:EPSG:'):
        code = crs.rpartition(':')[-1]
        crs = 'EPSG:%s' % code
    elif crs.startswith('EPSG'):
        pass
    elif crs in cfg.published_CRSs:
        pass
    else:
        raise WCS2Exception("Not a CRS: %s" % crs,
                            WCS2Exception.NOT_A_CRS,
                            locator=crs,
                            valid_keys=list(cfg.published_CRSs))
    return crs


def get_coverage_data(request, styles, qprof):
    # pylint: disable=too-many-locals, protected-access

    cfg = get_config()
    qprof.start_event("setup")
    layer_name = request.coverage_id
    layer = cfg.layer_index.get(layer_name)
    if not layer or not layer.wcs:
        raise WCS2Exception("Invalid coverage: %s" % layer_name,
                            WCS2Exception.NO_SUCH_COVERAGE,
                            locator="COVERAGE parameter",
                            valid_keys=list(cfg.layer_index))

    #
    # CRS handling
    #

    native_crs = layer.native_CRS
    subsetting_crs = uniform_crs(cfg, request.subsetting_crs or native_crs)
    if subsetting_crs not in cfg.published_CRSs:
        raise WCS2Exception("Invalid subsettingCrs: %s" % subsetting_crs,
                            WCS2Exception.SUBSETTING_CRS_NOT_SUPPORTED,
                            locator=subsetting_crs,
                            valid_keys=list(cfg.published_CRSs))

    output_crs = uniform_crs(cfg, request.output_crs or subsetting_crs or native_crs)

    if output_crs not in cfg.published_CRSs:
        raise WCS2Exception("Invalid outputCrs: %s" % output_crs,
                            WCS2Exception.OUTPUT_CRS_NOT_SUPPORTED,
                            locator=output_crs,
                            valid_keys=list(cfg.published_CRSs))

    #
    # Subsetting/Scaling
    #

    scaler = WCSScaler(layer, subsetting_crs)
    times = layer.ranges.times

    subsets = request.subsets

    if len(subsets) != len(set(subset.dimension.lower() for subset in subsets)):
        dimensions = [subset.dimension.lower() for subset in subsets]
        duplicate_dimensions = [
            item
            for item, count in collections.Counter(dimensions).items()
            if count > 1
        ]

        raise WCS2Exception("Duplicate dimension%s: %s" % (
                                's' if len(duplicate_dimensions) > 1 else '',
                                ', '.join(duplicate_dimensions)
                            ),
                            WCS2Exception.INVALID_SUBSETTING,
                            locator=','.join(duplicate_dimensions)
                            )

    for subset in subsets:
        dimension = subset.dimension.lower()
        if dimension == 'time':
            if isinstance(subset, Trim):
                if "," in subset.high:
                    raise WCS2Exception(
                        "Subsets can only contain 2 elements - the lower and upper bounds. For arbitrary date lists, use WCS1",
                        WCS2Exception.INVALID_SUBSETTING,
                        locator="time")
                if layer.time_resolution.is_subday():
                    low = parse(subset.low) if subset.low is not None else None
                    low = default_to_utc(low)
                    high = parse(subset.high) if subset.high is not None else None
                    high = default_to_utc(high)
                else:
                    low = parse(subset.low).date() if subset.low is not None else None
                    high = parse(subset.high).date() if subset.high is not None else None
                if low is not None:
                    times = [
                        time for time in times
                        if time >= low
                    ]
                if high is not None:
                    times = [
                        time for time in times
                        if time <= high
                    ]
            elif isinstance(subset, Slice):
                point = parse(subset.point).date()
                times = [point]
        else:
            try:
                if isinstance(subset, Trim):
                    scaler.trim(dimension, subset.low, subset.high)
                elif isinstance(subset, Slice):
                    scaler.slice(dimension, subset.point)
            except WCSScalerUnknownDimension:
                raise WCS2Exception('Invalid subsetting axis %s' % subset.dimension,
                                WCS2Exception.INVALID_AXIS_LABEL,
                                locator=subset.dimension)

    #
    # Transform spatial extent to native CRS.
    #
    scaler.to_crs(output_crs)

    #
    # Scaling
    #

    scales = request.scales
    if len(scales) != len(set(subset.axis.lower() for subset in scales)):
        axes = [subset.axis.lower() for subset in scales]
        duplicate_axes = [
            item
            for item, count in collections.Counter(axes).items()
            if count > 1
        ]
        raise WCS2Exception('Duplicate scales for ax%ss: %s' % (
                                'i' if len(duplicate_axes) == 1 else 'e',
                                ', '.join(duplicate_axes)
                            ),
                            WCS2Exception.INVALID_SCALE_FACTOR,
                            locator=','.join(duplicate_axes)
                            )

    for scale in scales:
        axis = scale.axis.lower()

        if axis in ('time', 'k'):
            raise WCS2Exception('Cannot scale axis %s' % scale.axis,
                                WCS2Exception.INVALID_SCALE_FACTOR,
                                locator=scale.axis
                                )
        else:
            if isinstance(scale, ScaleAxis):
                scaler.scale_axis(axis, scale.factor)
            elif isinstance(scale, ScaleSize):
                scaler.scale_size(axis, scale.size)
            elif isinstance(scale, ScaleExtent):
                scaler.scale_extent(axis, scale.low, scale.high)

    #
    # Rangesubset
    #

    band_labels = layer.band_idx.band_labels()
    if request.range_subset:
        bands = []
        for range_subset in request.range_subset:
            if isinstance(range_subset, str):
                if range_subset not in band_labels:
                    raise WCS2Exception('No such field %s' % range_subset,
                                WCS2Exception.NO_SUCH_FIELD,
                                locator=range_subset,
                                valid_keys=band_labels
                                )
                bands.append(range_subset)
            else:
                if range_subset.start not in band_labels:
                    raise WCS2Exception('No such field %s' % range_subset.start,
                                        WCS2Exception.ILLEGAL_FIELD_SEQUENCE,
                                        locator=range_subset.start,
                                        valid_keys = band_labels)
                if range_subset.end not in band_labels:
                    raise WCS2Exception('No such field %s' % range_subset.end,
                                        WCS2Exception.ILLEGAL_FIELD_SEQUENCE,
                                        locator=range_subset.end,
                                        valid_keys = band_labels)

                start = band_labels.index(range_subset.start)
                end = band_labels.index(range_subset.end)
                bands.extend(band_labels[start:(end + 1) if end > start else (end - 1)])
    # Uncomment to restore original styles parameter hack.
    #
    # elif styles:
    #     bands = get_bands_from_styles(styles, layer, version=2)
    #     if not bands:
    #         bands = band_labels
    else:
        bands = band_labels

    #
    # Format handling
    #

    if not request.format:
        fmt = cfg.wcs_formats_by_name[layer.native_format]
    else:
        try:
            fmt = cfg.wcs_formats_by_mime[request.format]
        except KeyError:
            raise WCS2Exception("Unsupported format: %s" % request.format,
                                WCS2Exception.INVALID_PARAMETER_VALUE,
                                locator="FORMAT",
                                valid_keys=list(cfg.wcs_formats_by_mime))

    if len(times) > 1 and not fmt.multi_time:
        raise WCS2Exception(
            "Format does not support multi-time datasets - "
            "either constrain the time dimension or choose a different format",
            WCS2Exception.INVALID_SUBSETTING,
            locator="FORMAT or SUBSET"
                            )
    affine = scaler.affine()
    geobox = GeoBox((scaler.size.y, scaler.size.x),
                             affine, cfg.crs(output_crs))

    stacker = DataStacker(layer,
                          geobox,
                          times,
                          bands=bands)
    qprof.end_event("setup")
    qprof.start_event("count-datasets")
    n_datasets = stacker.n_datasets()
    qprof.end_event("count-datasets")
    qprof["n_datasets"] = n_datasets

    try:
        layer.resource_limits.check_wcs(n_datasets,
                                              geobox.height, geobox.width,
                                              sum(layer.band_idx.dtype_size(b) for b in bands),
                                              len(times)
                                       )
    except ResourceLimited as e:
        if e.wcs_hard or not layer.low_res_product_names:
            raise WCS2Exception(
                f"This request processes too much data to be served in a reasonable amount of time. ({e}) "
                + "Please reduce the bounds of your request and try again.")
        stacker.resource_limited = True
        qprof["resource_limited"] = str(e)

    if n_datasets == 0:
        raise WCS2Exception("The requested spatio-temporal subsets return no data.",
                        WCS2Exception.INVALID_SUBSETTING,
                        http_response=404)

    qprof.start_event("fetch-datasets")
    datasets = stacker.datasets()
    qprof.end_event("fetch-datasets")
    if qprof.active:
        qprof["datasets"] = {
            str(q): [str(i) for i in ids]
            for q, ids in stacker.dsids().items()
        }
    qprof.start_event("load-data")
    output = stacker.data(datasets, skip_corrections=True)
    qprof.end_event("load-data")

    # Clean extent flag band from output
    raw_bands = [layer.band_idx.locale_band(b) for b in bands]
    for k, v in output.data_vars.items():
        if k not in raw_bands:
            output = output.drop_vars([k])

    #
    # TODO: configurable
    #
    if fmt.mime == 'image/geotiff':
        output = fmt.renderer(request.version)(request, output, output_crs,
                              layer, scaler.size.x, scaler.size.y, affine)

    else:
        output = fmt.renderer(request.version)(request, output, output_crs)

    headers = {
        "Content-Type": fmt.mime,
        'content-disposition': f'attachment; filename={request.coverage_id}.{fmt.extension}',
    }
    headers.update(layer.resource_limits.wcs_cache_rules.cache_headers(n_datasets))
    return output, headers


def get_tiff(request, data, crs, product, width, height, affine):
    """Uses rasterio MemoryFiles in order to return a streamable GeoTiff response"""
    # Does not support multi-time dimension data - is this even possible in GeoTiff?
    supported_dtype_map = {
        'uint8': 1,
        'int8': 2,
        'uint16': 3,
        'int16': 4,
        'uint32': 5,
        'int32': 6,
        'float32': 7,
        'float64': 8,
        'complex': 10,
        'complex64': 11,
        'complex128': 12,
    }

    dtype_list = [data[array].dtype for array in data.data_vars]
    dtype = str(max(dtype_list, key=lambda d: supported_dtype_map[str(d)]))

    # TODO: convert other parameters as-well
    gtiff = request.geotiff_encoding_parameters
    cfg = get_config()

    if len(data.time) > 1:
        raise WCS2Exception("Multiple time slices not supported by GeoTIFF format")
    data = data.squeeze(dim="time", drop=True)
    data = data.astype(dtype)
    nodata = 0
    for band in data.data_vars:
        nodata = product.band_idx.nodata_val(band)
    with MemoryFile() as memfile:
        # pylint: disable=protected-access, bad-continuation

        kwargs = {}
        if gtiff.tile_width is not None:
            kwargs['blockxsize'] = gtiff.tile_width
        if gtiff.tile_height is not None:
            kwargs['blockysize'] = gtiff.tile_height

        if gtiff.predictor:
            predictor = gtiff.predictor.lower()
            if predictor == 'horizontal':
                kwargs['predictor'] = 2
            elif predictor == 'floatingpoint':
                kwargs['predictor'] = 3
        elif dtype == "float64":
                kwargs["predictor"] = 3
        else:
            kwargs["predictor"] = 2

        with memfile.open(
            driver="GTiff",
            width=width,
            height=height,
            count=len(data.data_vars),
            transform=affine,
            crs=crs,
            nodata=nodata,
            tiled=gtiff.tiling if gtiff.tiling is not None else True,
            compress=gtiff.compression.lower() if gtiff.compression else "lzw",
            interleave=gtiff.interleave or "band",
            dtype=dtype, **kwargs) as dst:
            for idx, band in enumerate(data.data_vars, start=1):
                dst.write(data[band].values, idx)
                dst.set_band_description(idx, product.band_idx.band_label(band))
                if cfg.wcs_tiff_statistics:
                    dst.update_tags(idx, STATISTICS_MINIMUM=data[band].values.min())
                    dst.update_tags(idx, STATISTICS_MAXIMUM=data[band].values.max())
                    dst.update_tags(idx, STATISTICS_MEAN=data[band].values.mean())
                    dst.update_tags(idx, STATISTICS_STDDEV=data[band].values.std())
        return memfile.read()


def get_netcdf(request, data, crs):
    # Cleanup dataset attributes for NetCDF export
    data.attrs["crs"] = crs
    for v in data.data_vars.values():
        v.attrs["crs"] = crs
        if "spectral_definition" in v.attrs:
            del v.attrs["spectral_definition"]
        if "flags_definition" in v.attrs:
            del v.attrs["flags_definition"]
    if "time" in data and "units" in data["time"].attrs:
        del data["time"].attrs["units"]

    # And export to NetCDF
    return data.to_netcdf()
